#!/usr/bin/python3

"""visualise_result_MEGAHIT.py: Visualise the binning result from on the MEGAHIT assembly graph.

Visualize the binning result by denoting coloured contigs in the assembly
graph according to their corresponding bins. You can visualise the initial
binning result obtained from an existing binning tool and the final binning
result obtained from GraphBin and compare.

"""

import csv
import logging
import os
import random
import re
import subprocess
import sys

from cogent3.parse.fasta import MinimalFastaParser
from graphbin.bidirectionalmap.bidirectionalmap import BidirectionalMap
from igraph import *

__author__ = "Vijini Mallawaarachchi"
__copyright__ = "Copyright 2023-2024, GraphBin-Tk Project"
__credits__ = ["Vijini Mallawaarachchi", "Anuradha Wickramarachchi", "Yu Lin"]
__license__ = "GPL-3.0"
__version__ = "1.0.0"
__maintainer__ = "Vijini Mallawaarachchi"
__email__ = "viji.mallawaarachchi@gmail.com"
__status__ = "Production/Stable"

# create logger
logger = logging.getLogger(f"GraphBin-Tk {__version__}")


def run(args):

    initial_binning_result = args.initial
    final_binning_result = args.final
    assembly_graph_file = args.graph
    contigs_file = args.contigs
    output_path = args.output
    prefix = args.prefix
    dpi = args.dpi
    width = args.width
    height = args.height
    vsize = args.vsize
    lsize = args.lsize
    margin = args.margin
    image_type = args.imgtype
    delimiter = args.delimiter

    logger.info(f"Welcome to binning result visualiser of GraphBin-Tk!")
    logger.info(
        f"This version of the visualiser makes use of the assembly graph produced by MEGAHIT which is based on the de Bruijn graph approach."
    )

    # Validate prefix
    # ---------------------------------------------------
    try:
        if prefix != "":
            if not prefix.endswith("_"):
                prefix = prefix + "_"
        else:
            prefix = ""

    except:
        logger.error(f"Please enter a valid string for prefix")
        logger.info(f"Exiting visualiser... Bye...!")
        sys.exit(1)

    # Format type if provided
    # ---------------------------------------------------
    if image_type.startswith("."):
        image_type = image_type[1:]

    # Check if output folder exists
    # ---------------------------------------------------

    # Handle for missing trailing forwardslash in output folder path
    if output_path[-1:] != "/":
        output_path = output_path + "/"

    # Create output folder if it does not exist
    if not os.path.isdir(output_path):
        subprocess.run(f"mkdir -p {output_path}", shell=True)

    logger.info(f"Assembly graph file: {assembly_graph_file}")
    logger.info(f"Initial binning results file: {initial_binning_result}")
    logger.info(f"Final binning results file: {final_binning_result}")
    logger.info(f"Final output path: {output_path}")
    logger.info(f"Image type: {image_type}")
    logger.info(f"Width of image: {width} pixels")
    logger.info(f"Height of image: {height} pixels")
    logger.info(f"Size of the vertices: {vsize} pt")
    logger.info(f"Size of the vertex labels: {lsize} pt")
    logger.info(f"Size of the margin: {margin} pt")
    logger.info(f"Delimiter: {delimiter}")

    # Get the number of bins from the initial binning result
    # ---------------------------------------------------

    try:
        all_bins_list = []
        n_bins = 0

        with open(initial_binning_result, mode="r") as csvfile:
            readCSV = csv.reader(csvfile, delimiter=delimiter)
            for row in readCSV:
                all_bins_list.append(row[1])

        bins_list = list(set(all_bins_list))
        bins_list.sort()

        n_bins = len(bins_list)
        logger.info(f"Number of bins available in initial binning result: {n_bins}")

    except:
        logger.error(
            f"Please make sure that the correct path to the initial binning result file is provided and it is having the correct format"
        )
        logger.info(f"Exiting visualiser... Bye...!")
        sys.exit(1)

    logger.info("Constructing the assembly graph...")

    # Get original contig IDs
    # -------------------------------

    original_contigs = {}
    contig_descriptions = {}

    for label, seq in MinimalFastaParser(contigs_file):
        name = label.split()[0]
        original_contigs[name] = seq
        contig_descriptions[name] = label

    ## Construct the assembly graph
    # -------------------------------

    node_count = 0

    graph_contigs = {}

    links = []

    my_map = BidirectionalMap()

    try:
        # Get links from .gfa file
        with open(assembly_graph_file, mode="r") as file:
            for line in file.readlines():
                line = line.strip()

                # Identify lines with link information
                if line.startswith("L"):
                    link = []

                    strings = line.split("\t")

                    start_1 = "NODE_"
                    end_1 = "_length"

                    link1 = int(
                        re.search("%s(.*)%s" % (start_1, end_1), strings[1]).group(1)
                    )

                    start_2 = "NODE_"
                    end_2 = "_length"

                    link2 = int(
                        re.search("%s(.*)%s" % (start_2, end_2), strings[3]).group(1)
                    )

                    link.append(link1)
                    link.append(link2)
                    links.append(link)

                elif line.startswith("S"):
                    strings = line.split()

                    start = "NODE_"
                    end = "_length"

                    contig_num = int(
                        re.search("%s(.*)%s" % (start, end), strings[1]).group(1)
                    )

                    my_map[node_count] = int(contig_num)

                    graph_contigs[contig_num] = strings[2]

                    node_count += 1

        logger.info("Total number of contigs available: {node_count}")

        contigs_map = my_map
        contigs_map_rev = my_map.inverse

        # Map original contig IDs to contig IDS of assembly graph

        graph_to_contig_map = BidirectionalMap()

        for (n, m), (n2, m2) in zip(graph_contigs.items(), original_contigs.items()):
            if m == m2:
                graph_to_contig_map[n] = n2

        graph_to_contig_map_rev = graph_to_contig_map.inverse

        # Create graph
        assembly_graph = Graph()

        # Add vertices
        assembly_graph.add_vertices(node_count)

        # Create list of edges
        edge_list = []

        for i in range(node_count):
            assembly_graph.vs[i]["id"] = i
            assembly_graph.vs[i]["label"] = str(contigs_map[i])
            assembly_graph.vs[i]["name"] = graph_to_contig_map[contigs_map[i]]

        # Iterate links
        for link in links:
            # Remove self loops
            if link[0] != link[1]:
                # Add edge to list of edges
                edge_list.append((contigs_map_rev[link[0]], contigs_map_rev[link[1]]))

        # Add edges to the graph
        assembly_graph.add_edges(edge_list)
        assembly_graph.simplify(multiple=True, loops=False, combine_edges=None)

    except:
        logger.error(
            "Please make sure that the correct path to the assembly graph file is provided."
        )
        logger.info("Exiting visualiser... Bye...!")
        sys.exit(1)

    # Get initial binning result
    # ----------------------------

    logger.info("Obtaining the initial binning result...")

    bins = [[] for x in range(n_bins)]

    try:
        with open(initial_binning_result, mode="r") as contig_bins:
            readCSV = csv.reader(contig_bins, delimiter=delimiter)
            for row in readCSV:
                contig_num = contigs_map_rev[int(graph_to_contig_map_rev[row[0]])]
                bin_num = bins_list.index(row[1])
                bins[bin_num].append(contig_num)

    except:
        logger.error(
            f"Please make sure that the correct path to the binning result file is provided and it is having the correct format"
        )
        logger.info(f"Exiting visualiser... Bye...!")
        sys.exit(1)

    # Get isolated vertices
    # -------------------------------------------------

    # Get isolated contigs with no neighbours
    isolated = [v.index for v in assembly_graph.vs if v.degree() == 0]
    logger.info(f"Total isolated contigs in the assembly graph: {len(isolated)}")

    # Get list of colours according to number of bins
    # -------------------------------------------------

    logger.info("Picking colours out of 22 supported colours...")

    my_colours = [
        "#e6194b",
        "#3cb44b",
        "#ffe119",
        "#4363d8",
        "#f58231",
        "#911eb4",
        "#46f0f0",
        "#f032e6",
        "#bcf60c",
        "#fabebe",
        "#008080",
        "#e6beff",
        "#9a6324",
        "#fffac8",
        "#800000",
        "#aaffc3",
        "#808000",
        "#ffd8b1",
        "#000075",
        "#808080",
        "#ffffff",
        "#000000",
    ]

    # Visualise the initial assembly graph
    # --------------------------------------

    logger.info(
        "Drawing and saving the assembly graph with the initial binning result..."
    )

    initial_out_fig_name = f"{output_path}{prefix}initial_binning_result.{image_type}"

    node_colours = []

    for i in assembly_graph.vs()["name"]:
        contig_num = contigs_map_rev[int(graph_to_contig_map_rev[i])]
        no_bin = True
        for j in range(n_bins):
            if contig_num in bins[j]:
                node_colours.append(my_colours[j])
                no_bin = False

        if no_bin:
            node_colours.append("grey")

    assembly_graph.vs["color"] = node_colours

    visual_style = {}

    # Set bbox and margin
    visual_style["bbox"] = (width, height)
    visual_style["margin"] = margin

    # Set vertex size
    visual_style["vertex_size"] = vsize

    # Set vertex lable size
    visual_style["vertex_label_size"] = lsize

    # Don't curve the edges
    visual_style["edge_curved"] = False

    # Set the layout
    my_layout = assembly_graph.layout_fruchterman_reingold()
    visual_style["layout"] = my_layout

    # Plot the graph
    plot(assembly_graph, initial_out_fig_name, **visual_style)

    # Get the final GraphBin binning result
    # ---------------------------------------

    logger.info(f"Obtaining the final GraphBin binning result...")

    bins = [[] for x in range(n_bins)]

    try:
        with open(final_binning_result, mode="r") as contig_bins:
            readCSV = csv.reader(contig_bins, delimiter=delimiter)
            for row in readCSV:
                contig_num = contigs_map_rev[int(graph_to_contig_map_rev[row[0]])]
                bin_num = bins_list.index(row[1])
                bins[bin_num].append(contig_num)

    except:
        logger.error(
            f"Please make sure that the correct path to the final binning result file is provided and it is having the correct format"
        )
        logger.info(f"Exiting visualiser... Bye...!")
        sys.exit(1)

    # Visualise the final assembly graph
    # ------------------------------------

    logger.info(
        f"Drawing and saving the assembly graph with the final GraphBin binning result..."
    )

    final_out_fig_name = (
        f"{output_path}{prefix}final_GraphBin_binning_result.{image_type}"
    )

    node_colours = []

    for i in assembly_graph.vs()["name"]:
        contig_num = contigs_map_rev[int(graph_to_contig_map_rev[i])]
        no_bin = True
        for j in range(n_bins):
            if contig_num in bins[j]:
                node_colours.append(my_colours[j])
                no_bin = False

        if no_bin:
            node_colours.append("grey")

    assembly_graph.vs["color"] = node_colours

    visual_style = {}

    # Set bbox and margin
    visual_style["bbox"] = (width, height)
    visual_style["margin"] = margin

    # Set vertex size
    visual_style["vertex_size"] = vsize

    # Set vertex lable size
    visual_style["vertex_label_size"] = lsize

    # Don't curve the edges
    visual_style["edge_curved"] = False

    # Set the layout
    visual_style["layout"] = my_layout

    # Plot the graph
    plot(assembly_graph, final_out_fig_name, **visual_style)

    logger.info(
        f"Visualization of the initial binning results can be found at {initial_out_fig_name}"
    )
    logger.info(
        f"Visualization of the final GraphBin binning results can be found at {final_out_fig_name}"
    )

    # Exit program
    # --------------

    logger.info(f"Thank you for using visualiser for GraphBin-Tk!")


def main(args):
    run(args)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""evaluate.py: Evaluate the binning results given a ground truth.

Evaluate the binning results given a ground truth and calculate the
precision, recall, F1-score and ARI of the provided binning result.

"""

import csv
import logging

import scipy.special
from tabulate import tabulate

__author__ = "Vijini Mallawaarachchi"
__copyright__ = "Copyright 2023-2024, GraphBin-Tk Project"
__credits__ = ["Vijini Mallawaarachchi", "Anuradha Wickramarachchi", "Yu Lin"]
__license__ = "GPL-3.0"
__version__ = "1.0.0"
__maintainer__ = "Vijini Mallawaarachchi"
__email__ = "viji.mallawaarachchi@gmail.com"
__status__ = "Production/Stable"

# create logger
logger = logging.getLogger(f"GraphBin-Tk {__version__}")


def run(args):

    # Get paths to binning result and ground truth
    binned_file = args.binned
    ground_truth_file = args.groundtruth
    delimiter = args.delimiter
    prefix = args.prefix
    output_path = args.output

    # Setup output path for log file
    fileHandler = logging.FileHandler(f"{output_path}/gbintk.log")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    fileHandler.setLevel(logging.DEBUG)
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)

    logger.info(f"Welcome to binning result evaluator of GraphBin-Tk!")

    logger.info(f"Starting binning results evaluation...")
    logger.info(f"Binning results file: {binned_file}")
    logger.info(f"Ground truth file: {ground_truth_file}")
    logger.info(f"Delimiter: {delimiter}")
    logger.info(f"Output path: {output_path}")

    # Validate prefix
    # ---------------------------------------------------
    if prefix != "":
        if not prefix.endswith("_"):
            prefix = prefix + "_"
    else:
        prefix = ""

    # Get the number of bins from the ground truth
    # ---------------------------------------------------------
    ground_truth_n_bins = 0

    all_ground_truth_bins_list = []

    with open(ground_truth_file, mode="r") as csvfile:
        readCSV = csv.reader(csvfile, delimiter=delimiter)
        for row in readCSV:
            all_ground_truth_bins_list.append(row[1])

    ground_truth_bins_list = list(set(all_ground_truth_bins_list))
    ground_truth_n_bins = len(ground_truth_bins_list)

    logger.info(f"Number of bins available in the ground truth: {ground_truth_n_bins}")

    # Get the ground truth
    # ----------------------------
    ground_truth_bins = [[] for x in range(ground_truth_n_bins)]

    ground_truth_count = 0
    ground_truth_bins_1 = {}

    with open(ground_truth_file, mode="r") as contig_bins:
        readCSV = csv.reader(contig_bins, delimiter=delimiter)

        for row in readCSV:
            ground_truth_count += 1
            contig = row[0]
            bin_num = ground_truth_bins_list.index(row[1])
            ground_truth_bins[bin_num].append(contig)
            ground_truth_bins_1[contig] = bin_num

    logger.info(
        f"Number of contigs available in the ground truth: {ground_truth_count}"
    )

    # Get the number of bins from the initial binning result
    # ---------------------------------------------------------
    n_bins = 0

    all_bins_list = []

    with open(binned_file, mode="r") as csvfile:
        readCSV = csv.reader(csvfile, delimiter=delimiter)

        for row in readCSV:
            all_bins_list.append(row[1])

    bins_list = list(set(all_bins_list))
    n_bins = len(bins_list)

    logger.info(f"Number of bins available in the binning result: {n_bins}")

    # Get initial binning result
    # ----------------------------
    bins = [[] for x in range(n_bins)]

    bins_1 = {}

    binned_count = 0
    binned_contigs = []

    with open(binned_file, mode="r") as contig_bins:
        readCSV = csv.reader(contig_bins, delimiter=delimiter)
        for row in readCSV:
            binned_count += 1
            contig = row[0]
            bin_num = bins_list.index(row[1])
            bins[bin_num].append(contig)
            bins_1[contig] = bin_num
            binned_contigs.append(contig)

    logger.info(f"Number of contigs available in the binning result: {binned_count}")

    # Determine precision, recall, F1-score and ARI for binning result
    # ------------------------------------------------------------------

    total_binned = 0

    bins_species = [[0 for x in range(ground_truth_n_bins)] for y in range(n_bins)]

    for i in bins_1:
        if i in ground_truth_bins_1:
            total_binned += 1
            bins_species[bins_1[i]][ground_truth_bins_1[i]] += 1

    logger.info(
        f"Number of contigs available in the binning result that are present in the ground truth: {total_binned}"
    )
    logger.info(
        f"Number of unbinned contigs from the ground truth: {ground_truth_count-total_binned}"
    )

    logger.info(f"Ground truth bin labels:")

    for i in range(len(ground_truth_bins_list)):
        logger.info(f"{i+1}, {ground_truth_bins_list[i]}")

    logger.info(f"KxS Matrix:\n{tabulate(bins_species)}")

    my_precision = getPrecision(bins_species, n_bins, ground_truth_n_bins, total_binned)
    my_recall = getRecall(
        bins_species,
        n_bins,
        ground_truth_n_bins,
        total_binned,
        (ground_truth_count - total_binned),
    )
    my_ari = getARI(bins_species, n_bins, ground_truth_n_bins, total_binned)
    my_f1 = getF1(my_precision, my_recall)

    logger.info(f"Evaluation Results:")
    logger.info(f"Precision = {my_precision*100}")
    logger.info(f"Recall = {my_recall*100}")
    logger.info(f"F1-score = {my_f1*100}")
    logger.info(f"ARI = {my_ari*100}")

    # Write results to output file
    # -----------------------------

    output_file_path = f"{output_path}/{prefix}evaluation_results.txt"

    with open(output_file_path, mode="w") as eval_res:
        eval_res.write(f"KxS Matrix:\n{tabulate(bins_species)}\n")
        eval_res.write(f"Evaluation Results:\n")
        eval_res.write(f"Precision = {my_precision*100}\n")
        eval_res.write(f"Recall = {my_recall*100}\n")
        eval_res.write(f"F1-score = {my_f1*100}\n")
        eval_res.write(f"ARI = {my_ari*100}\n")

    logger.info(f"Evaluation results can be found in {output_file_path}")

    # Exit program
    # --------------

    logger.info(f"Thank you for using binning results evaluator for GraphBin-Tk!")


# Functions to determine precision, recall, F1-score and ARI
# ------------------------------------------------------------


# Get precicion
def getPrecision(mat, k, s, total):
    sum_k = 0
    for i in range(k):
        max_s = 0
        for j in range(s):
            if mat[i][j] > max_s:
                max_s = mat[i][j]
        sum_k += max_s
    return sum_k / total


# Get recall
def getRecall(mat, k, s, total, unclassified):
    sum_s = 0
    for i in range(s):
        max_k = 0
        for j in range(k):
            if mat[j][i] > max_k:
                max_k = mat[j][i]
        sum_s += max_k
    return sum_s / (total + unclassified)


# Get ARI
def getARI(mat, k, s, N):
    t1 = 0
    for i in range(k):
        sum_k = 0
        for j in range(s):
            sum_k += mat[i][j]
        t1 += scipy.special.binom(sum_k, 2)

    t2 = 0
    for i in range(s):
        sum_s = 0
        for j in range(k):
            sum_s += mat[j][i]
        t2 += scipy.special.binom(sum_s, 2)

    t3 = t1 * t2 / scipy.special.binom(N, 2)

    t = 0
    for i in range(k):
        for j in range(s):
            t += scipy.special.binom(mat[i][j], 2)

    ari = (t - t3) / ((t1 + t2) / 2 - t3)
    return ari


# Get F1-score
def getF1(prec, recall):
    return 2 * prec * recall / (prec + recall)


def main(args):
    run(args)


if __name__ == "__main__":
    main()



"""gbintk (GraphBin-Tk): Assembly graph-based metagenomic binning toolkit"""

__author__ = "Vijini Mallawaarachchi"
__copyright__ = "Copyright 2023-2024, GraphBin-Tk Project"
__credits__ = ["Vijini Mallawaarachchi", "Anuradha Wickramarachchi", "Yu Lin"]
__license__ = "GPL-3.0"
__version__ = "1.0.0"
__maintainer__ = "Vijini Mallawaarachchi"
__email__ = "viji.mallawaarachchi@gmail.com"
__status__ = "Production/Stable"



#!/usr/bin/python3

"""prep_result.py: Format the initial binning result from an existing binning tool.

Format the initial binning result from an existing binning tool in the .csv format
with contig ID and bin ID. Contigs are numbered starting from 0 and bins are 
numbered starting from 1.

"""


import csv
import logging
import os
import subprocess
import sys

from cogent3.parse.fasta import MinimalFastaParser

__author__ = "Vijini Mallawaarachchi"
__copyright__ = "Copyright 2023-2024, GraphBin-Tk Project"
__credits__ = ["Vijini Mallawaarachchi", "Anuradha Wickramarachchi", "Yu Lin"]
__license__ = "GPL-3.0"
__version__ = "1.0.0"
__maintainer__ = "Vijini Mallawaarachchi"
__email__ = "viji.mallawaarachchi@gmail.com"
__status__ = "Production/Stable"

# create logger
logger = logging.getLogger(f"GraphBin-Tk {__version__}")


def run(args):

    # Get arguments
    assembler = args.assembler
    contig_bins_folder = args.resfolder
    prefix = args.prefix
    delimiter = args.delimiter
    output_path = args.output

    # Setup output path for log file
    fileHandler = logging.FileHandler(f"{output_path}/gbintk.log")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    fileHandler.setLevel(logging.DEBUG)
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)

    logger.info(f"Welcome to binning results formatter of GraphBin-Tk!")

    logger.info(f"Formatting binning results...")
    logger.info(f"Binning results folder: {contig_bins_folder}")
    logger.info(f"Prefix: {prefix}")
    logger.info(f"Delimiter: {delimiter}")
    logger.info(f"Output path: {output_path}")

    # Get list of files in the folder path of binning result.
    files_list = os.listdir(contig_bins_folder)

    # Check if folder path of binning result is empty.
    # ---------------------------------------------------
    if len(files_list) == 0:
        logger.error(
            "Folder containing the initial binning result is empty. Please enter a valid path to the folder."
        )
        logger.info("Exiting gbintk prepare... Bye...!")
        sys.exit(1)

    # Check if binning result folder contains fasta files.
    # ---------------------------------------------------
    isFasta = False

    files = []

    for myfile in files_list:

        # Full path of the file
        filepath = os.path.join(contig_bins_folder, myfile)

        if os.path.isfile(filepath) and myfile.lower().endswith((".fasta", ".fa", ".fna")):
            isFasta = True
            files.append(myfile)

    if not isFasta:
        logger.error(
            f"Folder containing the initial binning result does not contain fasta files (.fasta, .fa or .fna)."
        )
        logger.info(f"Exiting gbintk prepare... Bye...!")
        sys.exit(1)

    logger.info(f"Found {len(files)} FASTA files")

    # Validate prefix
    # ---------------------------------------------------
    if prefix != "":
        if not prefix.endswith("_"):
            prefix = prefix + "_"
    else:
        prefix = ""

    # Format binning results.
    # ---------------------------------------------------

    logger.info(f"Formatting initial binning results")

    contig_bins = []

    for bin_file in files:
        if bin_file.lower().endswith((".fasta", ".fa", ".fna")):
            for contig_name, _ in MinimalFastaParser(
                f"{contig_bins_folder}/{bin_file}"
            ):
                bin_name = ".".join(bin_file.split(".")[:-1])
                if assembler == "megahit":
                    line = ["_".join(contig_name.split("_")[0:2]), bin_name]
                else:
                    line = [contig_name, bin_name]
                contig_bins.append(line)

    # Write binning results to output file.
    # ---------------------------------------------------

    logger.info(f"Writing initial binning results to output file")

    with open(
        f"{output_path}/{prefix}initial_contig_bins.csv", mode="w"
    ) as contig_bins_file:
        contig_writer = csv.writer(
            contig_bins_file,
            delimiter=delimiter,
            quotechar='"',
            quoting=csv.QUOTE_MINIMAL,
        )

        for row in contig_bins:
            contig_writer.writerow(row)

    logger.info(
        f"Formatted initial binning results can be found at {contig_bins_file.name}"
    )

    # Exit program
    # --------------

    logger.info("Thank you for using binning results formatter for GraphBin-Tk!")


def main(args):
    run(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import logging
from collections import OrderedDict
from typing import Mapping, Optional

import click

__author__ = "Vijini Mallawaarachchi"
__copyright__ = "Copyright 2023-2024, GraphBin-Tk Project"
__credits__ = ["Vijini Mallawaarachchi", "Anuradha Wickramarachchi", "Yu Lin"]
__license__ = "GPL-3.0"
__version__ = "1.0.0"
__maintainer__ = "Vijini Mallawaarachchi"
__email__ = "viji.mallawaarachchi@gmail.com"
__status__ = "Production/Stable"


# Setup logger
# ---------------------------------------------------

logger = logging.getLogger(f"GraphBin-Tk {__version__}")
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
consoleHeader = logging.StreamHandler()
consoleHeader.setFormatter(formatter)
consoleHeader.setLevel(logging.INFO)
logger.addHandler(consoleHeader)


class OrderedGroup(click.Group):
    """custom group class to ensure help function returns commands in desired order.
    class is adapted from Максим Стукало's answer to
    https://stackoverflow.com/questions/47972638/how-can-i-define-the-order-of-click-sub-commands-in-help
    """

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[Mapping[str, click.Command]] = None,
        **kwargs,
    ):
        super().__init__(name, commands, **kwargs)
        #: the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.commands


@click.group(
    cls=OrderedGroup, context_settings=dict(help_option_names=["-h", "--help"])
)
@click.version_option(__version__, "-v", "--version", is_flag=True)
def main():
    """gbintk (GraphBin-Tk): Assembly graph-based metagenomic binning toolkit"""
    pass


_assembler = click.option(
    "--assembler",
    help="name of the assembler used (SPAdes, MEGAHIT or Flye)",
    type=click.Choice(["spades", "megahit", "flye"], case_sensitive=False),
    required=True,
)
_graph = click.option(
    "--graph",
    help="path to the assembly graph file",
    type=click.Path(exists=True),
    required=True,
)
_contigs = click.option(
    "--contigs",
    help="path to the contigs file",
    type=click.Path(exists=True),
    required=True,
)
_paths = click.option(
    "--paths",
    help="path to the contigs.paths (metaSPAdes) or assembly.info (metaFlye) file",
    type=click.Path(exists=True),
    required=False,
)
_abundance = click.option(
    "--abundance",
    help="path to the abundance file",
    type=click.Path(exists=True),
    required=True,
)
_binned = click.option(
    "--binned",
    help="path to the .csv file with the initial binning output from an existing tool",
    type=click.Path(exists=True),
    required=True,
)
_output = click.option(
    "--output",
    help="path to the output folder",
    type=click.Path(dir_okay=True, writable=True, readable=True),
    required=True,
)
_prefix = click.option(
    "--prefix",
    help="prefix for the output file",
    type=str,
    default="",
    required=False,
)
_delimiter = click.option(
    "--delimiter",
    help="delimiter for input/output results. Supports a comma and a tab.",
    type=click.Choice(["comma", "tab"], case_sensitive=False),
    default="comma",
    show_default=True,
    required=False,
)
_nthreads = click.option(
    "--nthreads",
    help="number of threads to use.",
    type=int,
    default=8,
    show_default=True,
    required=False,
)


_click_command_opts = dict(
    no_args_is_help=True, context_settings={"show_default": True}
)


# Main GraphBin
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_assembler
@_graph
@_contigs
@_paths
@_binned
@_output
@_prefix
@click.option(
    "--max_iteration",
    help="maximum number of iterations for label propagation algorithm",
    type=int,
    default=100,
    show_default=True,
    required=False,
)
@click.option(
    "--diff_threshold",
    help="difference threshold for label propagation algorithm",
    type=click.FloatRange(0, 1),
    default=0.1,
    show_default=True,
    required=False,
)
@_delimiter
def graphbin(
    assembler,
    graph,
    contigs,
    paths,
    binned,
    output,
    prefix,
    max_iteration,
    diff_threshold,
    delimiter,
):
    """GraphBin: Refined Binning of Metagenomic Contigs using Assembly Graphs"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info("Running GraphBin...")

    from graphbin import graphbin_Flye, graphbin_MEGAHIT, graphbin_SPAdes

    # Make args class
    class GraphBinArgsObj:
        def __init__(
            self,
            assembler,
            graph,
            contigs,
            paths,
            binned,
            output,
            prefix,
            max_iteration,
            diff_threshold,
            delimiter,
        ):
            self.assembler = assembler
            self.graph = graph
            self.contigs = contigs
            self.paths = paths
            self.binned = binned
            self.output = output
            self.prefix = prefix
            self.max_iteration = max_iteration
            self.diff_threshold = diff_threshold
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = GraphBinArgsObj(
        assembler,
        graph,
        contigs,
        paths,
        binned,
        output,
        prefix,
        max_iteration,
        diff_threshold,
        delimiter,
    )

    # Run GraphBin
    # ---------------------------------------------------
    if assembler.lower() == "flye":
        graphbin_Flye.main(args)
    if assembler.lower() == "megahit":
        graphbin_MEGAHIT.main(args)
    if assembler.lower() == "spades":
        graphbin_SPAdes.main(args)


# Main GraphBin2
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_assembler
@_graph
@_contigs
@_paths
@click.option(
    "--abundance",
    help="path to the abundance file",
    type=click.Path(exists=True),
    required=True,
)
@_binned
@_output
@_prefix
@click.option(
    "--depthb",
    help="maximum depth for the breadth-first-search.",
    type=int,
    default=5,
    show_default=True,
    required=False,
)
@click.option(
    "--threshold",
    help="threshold for determining inconsistent vertices.",
    type=float,
    default=1.5,
    show_default=True,
    required=False,
)
@_delimiter
@_nthreads
def graphbin2(
    assembler,
    graph,
    contigs,
    paths,
    abundance,
    binned,
    output,
    prefix,
    depthb,
    threshold,
    delimiter,
    nthreads,
):
    """GraphBin2: Refined and Overlapped Binning of Metagenomic Contigs Using Assembly Graphs"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info(f"Running GraphBin2...")

    from graphbin2 import graphbin2_Flye, graphbin2_MEGAHIT, graphbin2_SPAdes

    # Make args class
    class GraphBin2ArgsObj:
        def __init__(
            self,
            assembler,
            graph,
            contigs,
            paths,
            abundance,
            binned,
            output,
            prefix,
            depth,
            threshold,
            delimiter,
            nthreads,
        ):
            self.assembler = assembler
            self.graph = graph
            self.contigs = contigs
            self.paths = paths
            self.abundance = abundance
            self.binned = binned
            self.output = output
            self.prefix = prefix
            self.depth = depth
            self.threshold = threshold
            self.nthreads = nthreads
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = GraphBin2ArgsObj(
        assembler,
        graph,
        contigs,
        paths,
        abundance,
        binned,
        output,
        prefix,
        depthb,
        threshold,
        delimiter,
        nthreads,
    )

    # Run GraphBin
    # ---------------------------------------------------
    if assembler.lower() == "flye":
        graphbin2_Flye.main(args)
    if assembler.lower() == "megahit":
        graphbin2_MEGAHIT.main(args)
    if assembler.lower() == "spades":
        graphbin2_SPAdes.main(args)


# Main MetaCoAG
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_assembler
@_graph
@_contigs
@_paths
@_abundance
@_output
@click.option(
    "--hmm",
    help="path to marker.hmm file.",
    default="auxiliary/marker.hmm",
    show_default=True,
    type=str,
    required=False,
)
@_prefix
@click.option(
    "--min_length",
    help="minimum length of contigs to consider for binning.",
    type=int,
    default=1000,
    show_default=True,
    required=False,
)
@click.option(
    "--p_intra",
    help="minimum probability of an edge matching to assign to the same bin.",
    type=click.FloatRange(0, 1),
    default=0.1,
    show_default=True,
    required=False,
)
@click.option(
    "--p_inter",
    help="maximum probability of an edge matching to create a new bin.",
    type=click.FloatRange(0, 1),
    default=0.01,
    show_default=True,
    required=False,
)
@click.option(
    "--d_limit",
    help="distance limit for contig matching.",
    type=int,
    default=20,
    show_default=True,
    required=False,
)
@click.option(
    "--depthlp",
    help="depth to consider for label propagation.",
    type=int,
    default=10,
    show_default=True,
    required=False,
)
@click.option(
    "--n_mg",
    help="total number of marker genes.",
    type=int,
    default=108,
    show_default=True,
    required=False,
)
@click.option(
    "--no_cut_tc",
    help="do not use --cut_tc for hmmsearch.",
    is_flag=True,
    default=False,
    show_default=True,
    required=False,
)
@click.option(
    "--mg_threshold",
    help="length threshold to consider marker genes.",
    type=click.FloatRange(0, 1, clamp=True),
    default=0.5,
    show_default=True,
    required=False,
)
@click.option(
    "--bin_mg_threshold",
    help="minimum fraction of marker genes that should be present in a bin.",
    type=click.FloatRange(0, 1, clamp=True),
    default=0.33333,
    show_default=True,
    required=False,
)
@click.option(
    "--min_bin_size",
    help="minimum size of a bin to output in base pairs (bp).",
    type=int,
    default=200000,
    show_default=True,
    required=False,
)
@_delimiter
@_nthreads
def metacoag(
    assembler,
    graph,
    contigs,
    paths,
    abundance,
    output,
    hmm,
    prefix,
    min_length,
    p_intra,
    p_inter,
    d_limit,
    depthlp,
    n_mg,
    no_cut_tc,
    mg_threshold,
    bin_mg_threshold,
    min_bin_size,
    delimiter,
    nthreads,
):
    """MetaCoAG: Binning Metagenomic Contigs via Composition, Coverage and Assembly Graphs"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info("Running MetaCoAG...")

    from metacoag import metacoag_runner

    # Make args class
    class MetaCoAGArgsObj:
        def __init__(
            self,
            assembler,
            graph,
            contigs,
            paths,
            abundance,
            output,
            hmm,
            prefix,
            min_length,
            p_intra,
            p_inter,
            d_limit,
            depth,
            n_mg,
            no_cut_tc,
            mg_threshold,
            bin_mg_threshold,
            min_bin_size,
            delimiter,
            nthreads,
        ):
            self.assembler = assembler
            self.graph = graph
            self.contigs = contigs
            self.paths = paths
            self.abundance = abundance
            self.output = output
            self.hmm = hmm
            self.prefix = prefix
            self.min_length = min_length
            self.p_intra = p_intra
            self.p_inter = p_inter
            self.d_limit = d_limit
            self.depth = depthlp
            self.n_mg = n_mg
            self.no_cut_tc = no_cut_tc
            self.mg_threshold = mg_threshold
            self.bin_mg_threshold = bin_mg_threshold
            self.min_bin_size = min_bin_size
            self.nthreads = nthreads
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = MetaCoAGArgsObj(
        assembler,
        graph,
        contigs,
        paths,
        abundance,
        output,
        hmm,
        prefix,
        min_length,
        p_intra,
        p_inter,
        d_limit,
        depthlp,
        n_mg,
        no_cut_tc,
        mg_threshold,
        bin_mg_threshold,
        min_bin_size,
        delimiter,
        nthreads,
    )

    # Run MetaCoAG
    # ---------------------------------------------------
    metacoag_runner.main(args)


# Main Prepare
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_assembler
@click.option(
    "--resfolder",
    help="path to the folder containing FASTA files for individual bins",
    type=click.Path(exists=True),
    required=True,
)
@_delimiter
@_prefix
@_output
def prepare(assembler, resfolder, delimiter, prefix, output):
    """Format the initial binning result from an existing binning tool"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info("Formatting the binning result...")
    from gbintk.support import prep_result

    # Make args class
    class PrepArgsObj:
        def __init__(self, assembler, resfolder, delimiter, prefix, output):
            self.assembler = assembler
            self.resfolder = resfolder
            self.prefix = prefix
            self.output = output
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = PrepArgsObj(assembler, resfolder, delimiter, prefix, output)

    # Run Results Formatter
    # ---------------------------------------------------
    prep_result.main(args)


# Main Visualise
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_assembler
@click.option(
    "--initial",
    help="path to the initial binning result",
    type=click.Path(exists=True),
    required=True,
)
@click.option(
    "--final",
    help="path to the final binning result",
    type=click.Path(exists=True),
    required=True,
)
@_graph
@_contigs
@_paths
@_output
@_prefix
@click.option(
    "--dpi",
    help="dpi value",
    type=int,
    default=300,
    show_default=True,
    required=False,
)
@click.option(
    "--width",
    help="width of the image in pixels",
    type=int,
    default=2000,
    show_default=True,
    required=False,
)
@click.option(
    "--height",
    help="height of the image in pixels",
    type=int,
    default=2000,
    show_default=True,
    required=False,
)
@click.option(
    "--vsize",
    help="size of the vertices",
    type=int,
    default=50,
    show_default=True,
    required=False,
)
@click.option(
    "--lsize",
    help="size of the vertex labels",
    type=int,
    default=8,
    show_default=True,
    required=False,
)
@click.option(
    "--margin",
    help="margin of the figure",
    type=int,
    default=50,
    show_default=True,
    required=False,
)
@click.option(
    "--imgtype",
    help="type of the image (png, eps, pdf, svg)",
    type=click.Choice(["png", "eps", "pdf", "svg"], case_sensitive=False),
    default="png",
    show_default=True,
    required=False,
)
@_delimiter
def visualise(
    assembler,
    initial,
    final,
    graph,
    contigs,
    paths,
    output,
    prefix,
    dpi,
    width,
    height,
    vsize,
    lsize,
    margin,
    imgtype,
    delimiter,
):
    """Visualise binning and refinement results"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info(
        f"Please note that only 22 distinct colours are supported for visualisation."
    )
    logger.info("Running Visualisation for binning and refinement results...")

    from gbintk.support import (
        visualise_result_Flye,
        visualise_result_MEGAHIT,
        visualise_result_SPAdes,
    )

    # Make args class
    class VizArgsObj:
        def __init__(
            self,
            assembler,
            initial,
            final,
            graph,
            contigs,
            paths,
            output,
            prefix,
            dpi,
            width,
            height,
            vsize,
            lsize,
            margin,
            imgtype,
            delimiter,
        ):
            self.assembler = assembler
            self.initial = initial
            self.final = final
            self.graph = graph
            self.contigs = contigs
            self.paths = paths
            self.output = output
            self.prefix = prefix
            self.dpi = dpi
            self.width = width
            self.height = height
            self.vsize = vsize
            self.lsize = lsize
            self.margin = margin
            self.imgtype = imgtype
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = VizArgsObj(
        assembler,
        initial,
        final,
        graph,
        contigs,
        paths,
        output,
        prefix,
        dpi,
        width,
        height,
        vsize,
        lsize,
        margin,
        imgtype,
        delimiter,
    )

    # Run Visualisation
    # ---------------------------------------------------
    if assembler.lower() == "flye":
        visualise_result_Flye.main(args)
    if assembler.lower() == "megahit":
        visualise_result_MEGAHIT.main(args)
    if assembler.lower() == "spades":
        visualise_result_SPAdes.main(args)


# Main Evaluate
# -------------------------------------------------------------------
@main.command(**_click_command_opts)
@_binned
@click.option(
    "--groundtruth",
    help="path to the .csv file with the ground truth",
    type=click.Path(exists=True),
    required=True,
)
@_delimiter
@_prefix
@_output
def evaluate(binned, groundtruth, delimiter, prefix, output):
    """Evaluate the binning results given a ground truth"""

    logger.info(
        f"Welcome to GraphBin-Tk: Assembly graph-based metagenomic binning toolkit!"
    )
    logger.info("Evaluating the binning results results...")
    from gbintk.support import evaluate

    # Make args class
    class EvalArgsObj:
        def __init__(self, binned, groundtruth, delimiter, prefix, output):
            self.binned = binned
            self.groundtruth = groundtruth
            self.prefix = prefix
            self.output = output
            self.delimiter = '\t' if delimiter == "tab" else ','

    # Make args object
    args = EvalArgsObj(binned, groundtruth, delimiter, prefix, output)

    # Run Evaluation
    # ---------------------------------------------------
    evaluate.main(args)
